/*
 *	EXPAND.C
 *	Phil Budne
 *
 *	Expansion for Ziv-Lempel / Wegman-Miller data compression
 */

# include <stdio.h>

# define DCTSIZ 2048
# define BUMP 503
# define BUFLEN 201

# define ALPSIZ 128
# define BITSWD 16
# define VECSIZ ((DCTSIZ + BITSWD - 1) / BITSWD)

# define CLRCOD 9215
# define NUMCOD 9216
# define NUMCHR 96
# define LOWCHR ' '
# define PFX1 '\033'
# define PFX2 '_'
# define SFX1 '\033'
# define SFX2 '/'

# define BREAK '\035'

extern char *zmalloc();		/* from ZMEM */
extern int chkbit(),setbit();	/* from CHKBIT */

extern inikbd(),rstkbd(),putkbd();
extern int chkkbd();
extern char getkbd();

extern inimdm(),rstmdm(),putmdm();
extern int chkmdm();
extern char getmdm();

char buffer[BUFLEN];		/* packet receive buffer */
char *dict[DCTSIZ];		/* string dictionary */
int count;			/* dict entries in use */
int prev;			/* # of last output entry */
int leng[DCTSIZ];		/* dict entry string lengths */

int rover;			/* NUR reap scan ptr */
unsigned vector[VECSIZ];	/* NUR bitvector */

int find() {			/* find a free entry */
	register int i;

	do {			/* find an entry to free */
		rover += BUMP;	/* bump rover */
		if( rover >= DCTSIZ ) rover -= DCTSIZ; /* mod if needed */
	} while( chkbit(rover) ); /* recently used? */

	zfree(dict[rover],leng[rover]);	/* free the string */
	return( rover );
} /* find */

init() {			/* set up the world */
	register i;

	for( i = 0; i < VECSIZ ; i++ )
		vector[i] = 0;

	rover = 0;		/* clear NUR rover */
	count = 0;		/* clear number of entries in use */
	prev = -1;		/* clear prev token */
} /* init */

char *mvchrs(ptr, entry)
register char *ptr;
int entry;
{
	register char *ep;
	register int i;

	ep = dict[entry];
	i = leng[entry];

	do *ptr++ = *ep++;
	while( --i > 0 );

	return( ptr );
} /* mvchrs */

doadd(str1,str2)		/* put str1 // str2 into dict */
register str1, str2;
{
	char *np;
	int j,l;

	str1 = str1 - ALPSIZ;
	str2 = str2 - ALPSIZ;
	l = ( str1 < 0 ? 1 : leng[str1] ) +
	    ( str2 < 0 ? 1 : leng[str2] );
	np = zmalloc(l);

	if( count < DCTSIZ ) j = count++; /* room in dict */
	else j = find();	/* no room, free an entry */

	dict[j] = np;		/* store ptr */
	leng[j] = l;		/* and length */

	if( str1 < 0 ) *np++ = (char) (str1+ALPSIZ);
	else np = mvchrs(np, str1);

	if( str2 < 0 ) *np = (char) (str2+ALPSIZ);
	else mvchrs(np, str2);

} /* doadd */

int expand(i)
int i;
{
	register j;

	if( i == CLRCOD ) {
		init();
		return( 1 );
	}

	if( i < ALPSIZ ) putkbd(i);
	else {
		j = i - ALPSIZ;		/* make dict entry */
/*		if( j >= count ) return( 0 ); /* not yet loaded? */
		kputs(dict[j],leng[j]);	/* output it */
		setbit(j);		/* mark as used recently */
	}

	if( prev != -1 ) doadd(prev,i);
	prev = i;
	return( 1 );
} /* expand */

dokbd() {			/* called when kbd ready */
	register char c;
	do {
		if( (c = getkbd()) == BREAK ) quit();
		else putmdm( c );
	}
	while( chkkbd() );
} /* dokbd */
	
char gm() {
	do if( chkkbd() ) dokbd(); /* service keyboard first */
	while( ! chkmdm() );	/* until we have a char on modem port */

	return( getmdm() & 0177 );
} /* gm */

# ifdef DOCRC

int process() {			/* process packet in buffer */
	char *bp;
	int crc;
	unsigned t;	

	prev = -1;
	crc = 0;
	bp = buffer;
	if( *bp == '\0' ) return( 0 );
	crc = *bp++ - LOWCHR;
	if( *bp == '\0' ) return( 0 );
	crc = crc * NUMCHR + (*bp++ - LOWCHR);

	t = 0;
	while( *bp != '\0' )
		t = t + (*bp++ - LOWCHR);
	t = t % NUMCOD;
	if( t != crc ) return( 0 );

	bp = &buffer[2];
	while( *bp != '\0' ) {
		t = *bp++ - LOWCHR;
		if( *bp == '\0' ) return( 0 ); /* short packet? */
		t = t * NUMCHR + (*bp++ - LOWCHR);
		if( ! expand(t) ) return( -1 );	/* restart!! */
	} /* buffer loop */
	return( 1 );
} /* process */

int getpack() {			/* get a packet into buffer */
	register char c;
	int state,count;
	char *bp;

	state = 0;
	count = BUFLEN-1;
	bp = buffer;

	for( ; ; ) {
		switch( state ) {
		case 0:		/* reading */
			c = gm();
			if( c == SFX1 ) {
				c = gm();
				if( c != SFX2 ) state = 3; /* error */
				else state = 2;	/* done */
			}
			else if( c < LOWCHR ) state = 3; /* error (nonprint) */
			else if( count-- < 0 ) state = 3; /* error (overrun) */
			else *bp++ = c;	/* store */
			break;

		case 2:		/* have packet */
			*bp++ = '\0';
			return( process() );

		case 3:		/* error -- looking for SFX1 */
			if( gm() == SFX1 ) state = 4;
			break;

		case 4:		/* error -- looking for SFX2 */
			if( (c = gm()) == SFX2 ) state = 5;
			else if( c != SFX1 ) state = 3;
			break;

		case 5:		/* error -- have seen suffix */
			return( -1 );

		} /* switch */
	} /* forever */
} /* getpack */

expmod() {
	int i;

	i = getpack();
	if( i == -1 ) putkbd('N');
	else if( i == 0 ) putkbd('0');
	else /* i == 1 */ putkbd('Y');
}
# endif

# ifndef DOCRC
expmod() {
	register char c,d;
	register int state;

	for( ; ; ) {
		switch( state ) {
		case 0:
			if( (c = gm()) == SFX1 ) state = 1;
			else state = 2;
			break;
		case 1:
			if( gm() == SFX2 ) return;
			putkbd( '\07' ); /* not suffix */
			state = 0;
			break;
		case 2:
			d = gm();
			if( d >= ' ' ) {
				expand( (c - ' ') * 96 + d - ' ' );
				state = 0;
			}
			else if( d == SFX1 ) state = 1;
			else state = 0;
			break;
		} /* switch */
	} /* for */
} /* expmod */
# endif

doit() {
	register char c, d;

	for( ; ; ) {
		if( (c = gm()) == PFX1 ) {
			if( (d = gm()) == PFX2 ) expmod();
			else {
				putkbd(c);
				putkbd(d);
			} /* failed on second */
		} /* first matched */
		else putkbd(c); /* failed on first */
	} /* for ever */
} /* doit */

quit() {
	rstkbd();		/* restore terminal driver */
	rstmdm();		/* restore modem driver */
	stats();		/* dumps stats */
	exit();			/* be gone */
} /* quit */

kputs(s,l)			/* put a string to kbd */
register char *s;
register int l;
{
	do putkbd( *s++ );
	while( --l > 0 );
} /* kputs */

main() {
	init();			/* init data structures */
	inimdm();		/* init modem device */
	inikbd();		/* init keyboard device */
	doit();			/* do work (never returns) */
} /* main */
